#include <string.h>
#include <stdio.h>

#include "types.h"
#include "header.h"
#include "files.h"
#include "sysdep.h"
#include "key.h"
#include "menu.h"

#define min(x,y) ((x)<(y)?(x):(y))

char dumpstr[256];

int colors=0,topslot;
extern int ascent,descent,wscreen,hscreen,gl_hchar,gl_wchar,hoffset,voffset;

void cursor (int x, int y)
/******* cursor
	invert a position on the screen, making or deleting the cursor.
*******/
{	
	XFillRectangle(display,twindow,xorgc,x,y-gl_hchar+1,
		gl_wchar,gl_hchar-1);
}

#define AGREEN 1

void outsofar (char *p, int attrib, int *x, int *y)
/******* outsofar
	help function for unscramble.
	output the string p wit attributes attrib at screen position
	x and y.
	mainly handles the attributes (like AGREEN=block on).
*******/
{
	int l=(int)strlen(p);	
	/* set graphics */
	if (attrib&AGREEN)
		XDrawImageString(display,twindow,invgc,*x,*y-descent,p,l);
	else XDrawImageString(display,twindow,gc,*x,*y-descent,p,l);
	/* update x */
	*x+=l*gl_wchar;
	*p=0;
}

int unscramble (char *from, char *end,
	int left, int length, int tabsize, int x, int y, int width)
/***** unscramble
	reformat the file content to printable screen characters,
	line at from, from left column, print the line and
	return number of printed characters. 
*****/
{	int count=0,total,i,adv,attrib=0;
	char p[256],*to;
	total=left+length;
	to=p;
	
	while ( (*from!=10) && (from<end) &&
				(count<left)) 
	{	if (*from==9)	count=(count/tabsize+1)*tabsize;
		else	count++;
		from++;
	}
	if (count>left)
		for (i=0; i<count-left; i++) *to++=' ';

	if (block.slot>=0 && (from>=block.start)&&(from<=block.end))
		attrib|=AGREEN;
	
	while ( (from<end) && (*from!=10) &&
				(count<total))
	{	if (block.slot>=0 && from==block.start) 
		{	*to=0; outsofar(p,attrib,&x,&y); to=p;
			attrib|=AGREEN;
		}
		if (block.slot>=0 && from==block.end)
		{	*to=0; outsofar(p,attrib,&x,&y); to=p;
			attrib&=~AGREEN;
		}	
		if (*from==9)
		{	adv=(count/tabsize+1)*tabsize-count;
			for (i=0; i<adv; i++) *to++=' ';
			from++;
			count+=adv;
		}
		else
			{	*to++=*from++; count++; }
	}
	*to=0;
	outsofar(p,attrib,&x,&y);
	if (x<width+hoffset)
	{	
		XFillRectangle(display,twindow,invgc,x,y-gl_hchar,
			width+hoffset-x,gl_hchar+1);
	}
	return count;
}

void redraw_line (char *line, filetyp *ft, int x, int y)
/***** redraw_line
	draw a line on screen plus cursor.
*****/
{	
	int h;
	unscramble(line,ft->fend,ft->left,ft->wl,ft->tabs,x,y,ft->w);
	if (line==ft->cursor.line)
	{	ft->cursor.x=find_pos(ft->cursor.line,ft->cursor.x,fend(ft),
			&(ft->cursor.adr),ft->tabs);
		if ((h=ft->cursor.x-ft->left)>=0 && (h<ft->left+ft->wl))
			if (cursoron) cursor(x+h*gl_wchar,y);
	}
}

void clip (int slot)
/***** clip
	set clipping to the specified slot (must be top window).
******/
{	
}

int find_slot (int whandle)
/****** find_slot
	see if this window is one of ours 
*******/
{	
	int slot;
	for (slot=MAXSLOTS-1; slot>=0; slot--)
		if (files[slot].flags.used &&
			(files[slot].whandle==whandle)) return slot;
	return -1;
}

void set_slider (char *line, filetyp *ft)
/***** set_slider
	 set vertical slider 
*****/
{	
}

void bline (int pos, char *s)
{
	XFillRectangle(display,bwindow,invgc,
		hoffset+pos*gl_wchar, hoffset,
		wscreen-(hoffset+pos*gl_wchar),gl_hchar);
	XDrawString(display,bwindow,gc,
	   hoffset+pos*gl_wchar,
	   hoffset+gl_hchar-descent-1,s,strlen(s));
}

void draw_scroll (filetyp *ft)
{	int w,h,l;
	w=vwidth; h=vheight-8-vwidth;
	l=(ft->topl-ft->start)*h/(ft->fend-ft->start+1);
	XMoveWindow(display,swindow,2,l+2);
}

void draw_arrows (void)
{	int d=vwidth/4;
	XDrawLine(display,vuwindow,gc,d,3*d,2*d,d);
	XDrawLine(display,vuwindow,gc,2*d,d,3*d,3*d);
	XDrawLine(display,vdwindow,gc,d,d,2*d,3*d);
	XDrawLine(display,vdwindow,gc,3*d,d,2*d,3*d);
}

extern char *recordp;
void draw_bline (void)
{	char string[256];
	int slot;
	filetyp *ft;
	slot=get_top();
	if (slot<0) return;
	ft=files+slot;
	sprintf(string,"%c %s %ld %s",ft->flags.updated?'*':' ',
		ft->name,
		(long)(ft->fend-ft->start)-
		(ft->gap.open?(long)(ft->gap.end-ft->gap.start):0l),
		recordp?"(Recording)":"");
	bline(0,string);
}	

void wredraw (int slot, int clip[4], Window w)
/***** wredraw 
	redraw one of our windows
******/
{	
	char *line,*newline;
	filetyp *ft;
	int i,x,y;

	if (slot<0) { clear_screen(); return; }
	ft=files+slot;
	/* draw all the lines */
	if (w==twindow || w==window)
	{	line=ft->topl;
		x=ft->xl; y=ft->y; clip[3]+=gl_hchar-1;
		for (i=0; i<ft->hl; i++)
		{	y+=gl_hchar;
			if (line>ft->fend) break;
			if (y>=clip[1]) redraw_line(line,ft,x,y);
			newline=next_line(line,ft);
			if (newline==line)
			{
				XFillRectangle(display,twindow,invgc,x,y+1,ft->w,ft->h);
				break;
			}
			else line=newline;
			if (y>clip[3]) break;
		}
	}
	if (w==bwindow || w==window) draw_bline();
	if (w==window) draw_scroll(ft);
	if (w==vuwindow || w==vdwindow || w==window) draw_arrows();
	
	XFlush(display);
	/* set slider */	
}

void redraw (int whandle, int p[4])
/***** redraw 
	redraw a window clipped in coordinates p 
******/
{	int slot;

	/* is it one of our windows? */
	slot=find_slot(whandle);
				
	/* if so,redraw it */
	if (slot>=0) wredraw(slot,p,window);
}

void refresh_top (void)
{
	int p[4];
	p[0]=p[1]=0; p[2]=wdesk; p[3]=hdesk;
	wredraw(topslot,p,window);
}

void draw_line (char *line, int slot, int ypos)
/****** draw_line
******/
{	
	filetyp *ft;
	ft=files+slot;
	
	/* clear the line */
	XFillRectangle(display,twindow,invgc,ft->xl,ypos-gl_hchar,
		ft->wl*gl_wchar,gl_hchar);
	
	/* redraw the line */
	redraw_line(line,ft,ft->xl,ypos);
}

void vscroll (int slot, int lines)
/********* vscroll
	scroll a file vertically lines lines
*******/
{	
	char *line,*l,*ll;
	filetyp *ft;
	int i;
	ft=files+slot;

	if (lines==0) return;
	/* special handling if scrolling only one line */
	if ((lines==1)||(lines==-1))
	{	
		/* that is, window topped */
		{	
			line=ft->topl;
			if (lines==1)
			{	l=line;
				line=next_line(line,ft);
				if (line==l) return;
				XCopyArea(display,twindow,twindow,gc,
					ft->xl,ft->y+gl_hchar,
					min(ft->wl*gl_wchar,ft->w),
					(ft->hl-1)*gl_hchar,
					ft->xl,ft->y);
				l=line;
				for (i=0; i<ft->hl-1; i++) 
				{	ll=next_line(l,ft);
					if (ll==l) 
					{	l=""; break; }
					l=ll;
				}
				draw_line(l,slot,ft->y+ft->hl*gl_hchar);
				XSync(display,0);
			}
			else
			{	l=line;
				line=prev_line(line,ft);
				if (line==l) return;
				XCopyArea(display,twindow,twindow,gc,
					ft->xl,ft->y,
					min(ft->wl*gl_wchar,ft->w),
					(ft->hl-1)*gl_hchar,
					ft->xl,ft->y+gl_hchar);
				draw_line(line,slot,ft->y+gl_hchar);
				l=line;
				for (i=0; i<ft->hl-1; i++) l=next_line(l,ft);
				XSync(display,0);
			}
			ft->topl=line;
		}
		if (params.scrollbar) draw_scroll(ft);
		return;
	}

	/* reset topline pointer */
	line=ft->topl;
	if (lines>0)
		for (i=0; i<lines; i++) 
		{	l=next_line(line,ft);
			if (l==line) break;
			line=l;
		}
	else
		for (i=0; i<-lines; i++) 
		{	l=prev_line(line,ft);
			if (l==line) break;
			line=l;
		}
	ft->topl=line;
	
	/* redraw the screen */
	refresh_top();
}

void hscroll (int slot, int columns)
/******** hscroll
	scroll a slot horiontally.
*********/
{	
	files[slot].left+=columns;
	if (files[slot].left<0) files[slot].left=0;
	refresh_top();
}

void wsize (int slot)
/****** wsize
	compute sizes of window, such as linewidth ... 
******/
{	
	int x,y,w,h;
	filetyp *ft=files+slot;
	x=0; y=0; w=twidth; h=theight;
	ft->x=x;
	ft->y=y;
	ft->xl=x+hoffset;
	ft->w=w;
	ft->h=h;
	ft->wl=(w-(ft->xl-x))/gl_wchar;
	ft->hl=(h-1)/gl_hchar;
}

void wsize_file (int whandle, int x, int y, int w, int h)
/****** wsize_file
	user resized the window.
******/
{	
	int slot;
	for (slot=0; slot<MAXSLOTS; slot++)
		if (files[slot].flags.used) wsize(slot);
	refresh_top();
}

void wtop_file (int whandle)
{	
	int slot;
	slot=find_slot(whandle);
	if (slot<0) return;
	topslot=slot;
	refresh_top();
}

void cycle_window (void)
{	
	int slot,i;
	slot=topslot;
	if (slot<0) return;
	i=slot+1;
	while (i!=slot)
	{	
		if (i>=MAXSLOTS) i=0;
		if (files[i].flags.used)
		{	
			topslot=i;
			break;
		}
		i++;
	}
	refresh_top();
}

int find_cursor (filetyp *ft)
/******** find_cursor
	find the cursorline on screen , given the file pointer.
*********/
{	
	int i;
	char *line=ft->topl,*cp=ft->cursor.line;
	for (i=0; i<ft->hl; i++)
		if (cp==line) return i;
		else line=next_line(line,ft); 
	return -1;
}

